from typing import Any, Dict

config: Dict[str, Any] = {
    "img_dir": "img",
    "size_small": (4.0, 3.0),
    "size_large": (8.0, 6.0),
    "dpi": 300,
    "thumbnail_scale": 0.5,
    "thumbnail_quality": 42,
}
"""Configuration used in :class:`nbfigtulz.figtools.save_fig`.

- ``img_dir``: The directory where the generated images are stored. This directory is not created by ``nbfigtulz`` and has to exist before :class:`nbfigtulz.figtools.save_fig` is called for the first time.
- ``size_small``: Default width and height of small images in units of inches.
- ``size_large``: Default width and height of large images in units of inches.
- ``dpi``: DPI used to store images.
- ``thumbnail_scale``: The thumbnail can be scaled w.r.t. the generated PNG to save space. Setting the scale to, e.g., 0.5 shrinks the PNG by a factor of 2.
- ``thumbnail_quality``: The image quality of the thumbnail, on a scale from 1 (worst) to 95 (best). 
"""



import base64
import io
from typing import Any, Optional, Tuple

from IPython.display import display, HTML
from PIL import Image


class Thumbnail:
    """Wrapper for a PNG image.

    :param bytes: Bytes of the PNG image.
    :param file_name: The file name.
    :param thumbnail_scale: Scaling factor for thumbnail.
    :param thumbnail_quality: Image quality of the thumbnail.
    :param bkg_color: Background color as RGB tuple.
    :param width: If not ``None`` the image is rescaled to the given width in units of pixels.
    :param print_compression: Print the compression rate of the thumbnail w.r.t. the PNG.
    """

    def __init__(
        self,
        bytes: Any,
        file_name: str,
        thumbnail_scale: float,
        thumbnail_quality: int,
        bkg_color: Tuple[int, int, int],
        width: Optional[int] = None,
        print_compression: bool = False,
    ):
        self.b64encoded = Thumbnail._bytes_to_decoded_b64(bytes)
        self.file_name = file_name
        self.width = width

        png = Image.open(io.BytesIO(bytes)).convert("RGBA")

        w, h = png.size
        size = int(w * thumbnail_scale), int(h * thumbnail_scale)

        quality = int(thumbnail_quality)
        if quality < 1:
            quality = 1
        elif quality > 95:
            quality = 95

        buffer = io.BytesIO()
        bkg = Image.new("RGBA", size, bkg_color)
        Image.alpha_composite(bkg, png.resize(size)).convert("RGB").save(
            buffer, format="JPEG", optimize=True, quality=quality
        )
        self.thumbnail = Thumbnail._bytes_to_decoded_b64(buffer.getvalue())

        if print_compression:
            rate = len(self.b64encoded) / len(self.thumbnail)
            print(f"Compression rate: {rate:.1f}")

    @staticmethod
    def _bytes_to_decoded_b64(bytes):
        return base64.b64encode(bytes).decode("ascii")

    def to_html(self, link: bool = True) -> str:
        """Renders the image into an HTML image tag.

        :param link: Whether or not to make the image tag a downloadable file link.
        :return: An HTML image tag.
        """
        img = "<img "
        if self.width:
            img += f"width={self.width} "
        img += f'src="data:image/jpeg;base64, {self.thumbnail}" />'

        if not link:
            return img

        link_temp = (
            f'<a download="{self.file_name}.png" '
            + f'href="data:image/png;base64, {self.b64encoded}">'
            + "{body}</a>"
        )
        return link_temp.format(body=img)

    def __repr__(self):
        display(HTML(self.to_html(link=True)))
        return self.file_name + ".png"



import enum
import functools
import io
import pathlib
from typing import Any, Callable, Dict, Optional, Tuple

import matplotlib as mpl
from IPython.display import HTML

from . import config as cfg
from . import thumbnail


class Size(enum.Enum):
    """Common image sizes.

    This is nothing but a convenient wrapper for the image sizes stored inside of :class:`nbfigtulz.config.config`.
    Possible values are :class:`Size.SMALL` and :class:`Size.LARGE`.
    """

    SMALL = 1
    """Refers to ``size_small`` in :class:`nbfigtulz.config.config`.
    """

    LARGE = 2
    """Refers to ``size_large`` in :class:`nbfigtulz.config.config`.
    """

    def get_size(self) -> Tuple[float, float]:
        """Returns the image size as a tuple of width and height.

        :return: The width and height as a tuple of floats in units of inches.
        """
        if self == Size.SMALL:
            return cfg["size_small"]
        if self == Size.LARGE:
            return cfg["size_large"]

        raise Exception("Unknown size")


class FigContext:
    """Context manager that temporally overwrites ``matplotlib.rcParams`` and the backend.

    :param backend: The new backend, defaults to ``"pgf"``.
    :param rcParams: The temporal ``rcParams``, defaults to ``lualatex`` for ``pgf.texsystem``.
    """

    def __init__(
        self, backend: str = "", rcParams: Optional[Dict[str, Any]] = None
    ) -> None:
        self.backend = backend if backend else "pgf"
        self.old_backend = mpl.get_backend()

        self.rcParams = {
            "font.family": "serif",
            "text.usetex": True,
            "pgf.texsystem": "lualatex",
            "pgf.rcfonts": False,
        }

        if rcParams:
            for k in rcParams:
                self.rcParams[k] = rcParams[k]

        self.old_rcParams = {k: mpl.rcParams[k] for k in self.rcParams}

    def __enter__(self):
        mpl.use(self.backend)
        for k in self.rcParams:
            mpl.rcParams[k] = self.rcParams[k]

    def __exit__(self, exc, *args, **kwargs) -> bool:
        mpl.use(self.old_backend)
        for k in self.rcParams:
            mpl.rcParams[k] = self.old_rcParams[k]

        return exc is None


def with_context(
    func: Optional[Callable] = None,
    rcParams: Optional[Dict[str, Any]] = None,
    preamble: Optional[Tuple[str, ...]] = None,
) -> Callable:
    """Wraps a function call inside the :class:`FigContext` context.

    :param func: A function.
    :param rcParams: Temporal ``rcParams`` for :class:`FigContext`.
    :param preamble: List of commands for ``rcParams['pgf.preamble']``.
    :return: The wrapped function.
    """
    if not rcParams:
        rcParams = dict()

    if preamble:
        key = "pgf.preamble"
        if key not in rcParams:
            rcParams[key] = ""
        else:
            rcParams[key] += "\n"

        rcParams["pgf.preamble"] += "\n".join(p for p in preamble)

    if callable(func):

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            with FigContext(rcParams=rcParams):
                return func(*args, **kwargs)

        return wrapped

    return functools.partial(with_context, rcParams=rcParams)


def save_fig(
    fig: Any,
    filename_base: str,
    resize: Any = Size.SMALL,
    *,
    suppress_pgf: bool = False,
    quiet: bool = False,
    thumbnail_scale: Optional[float] = None,
    tight_layout: bool = True,
    thumbnail_bkg: Optional[Tuple[int, int, int]] = None,
    **kwargs,
) -> thumbnail.Thumbnail:
    """The provided figure is stored to disk in the PNG and PGF (optional) format.

    :param fig: A ``matplotlib.pyplot.figure`` instance.
    :param filename_base: Base name (w/o file type suffix) of the PNG and PGF file.
    :param resize: If not ``None`` this will resize the figure. Pass anything with a ``get_size() -> Tuple[float, float]`` member function (e.g., :class:`Size`) or a tuple of two floats. Those floats are interpreted as the new width and height in units of inches, respectively.
    :param suppress_pgf: Suppress the generation of the PGF file.
    :param quiet: Do not print the FQNs of the generated files.
    :param thumbnail_scale: If not ``None`` this overwrites the default thumbnail scaling in :class:`nbfigtulz.config.config`.
    :param tight_layout: Require tight layout.
    :param thumbnail_bkg: Background color of thumbnail as RGB tuple. (Default: ``(255, 255, 255)`` aka white.)
    :param kwargs: Arguments passed to ``matplotlib.pyplot.savefig``.
    :return: The rendered PNG image.
    """
    if "dpi" not in kwargs:
        kwargs["dpi"] = cfg["dpi"]

    if resize:
        size = resize.get_size() if hasattr(resize, "get_size") else resize
        fig.set_size_inches(size)

    ftypes = [
        "png",
    ]
    if mpl.get_backend() == "pgf":
        if not suppress_pgf:
            ftypes.append("pgf")

        if tight_layout:
            fig.tight_layout()

    for ftype in ftypes:
        filename = str(pathlib.Path(cfg["img_dir"]) / f"{filename_base}.{ftype}")
        fig.savefig(filename, **kwargs)

        if not quiet:
            print(filename)

    img_bytes = io.BytesIO()
    fig.savefig(img_bytes, format="png", **kwargs)
    img_bytes.seek(0)

    width, _ = fig.get_size_inches() * 100

    if not thumbnail_scale:
        thumbnail_scale = cfg["thumbnail_scale"]

    if thumbnail_bkg is None:
        thumbnail_bkg = (255, 255, 255)

    return thumbnail.Thumbnail(
        img_bytes.read(),
        filename_base,
        width=width,
        thumbnail_scale=thumbnail_scale,
        thumbnail_quality=cfg["thumbnail_quality"],
        bkg_color=thumbnail_bkg,
    )


def img_grid(
    images: Tuple[thumbnail.Thumbnail, ...],
    *,
    n_columns: Optional[int] = None,
    width: Optional[int] = None,
) -> HTML:
    """Arranges images in a grid.

    :param images: List of images.
    :param n_columns: Number of columns, defaults to number of images.
    :param width: If not ``None`` the width of the grid in units of pixels.
    :return: The image grid.
    """
    if n_columns is None:
        n_columns = len(images)

    cells = [img.to_html() if img else "" for img in images]

    n_rows = len(cells) // n_columns
    if n_rows * n_columns < len(cells):
        n_rows += 1

    rows = []
    for i in range(n_rows):
        start = i * n_columns
        end = start + n_columns
        row = [
            f'<td style="text-align:center">{cell}</td>' for cell in cells[start:end]
        ]
        rows.append("".join(row))

    body = "".join([f'<tr style="background-color: white">{row}</tr>' for row in rows])

    table_temp = "<table "
    if width:
        table_temp += f'style="width: {width}px"'
    table_temp += ">{body}</table>"

    return HTML(table_temp.format(body=body))



"""
- :class:`nbfigtulz.config.config` as ``nbfigtulz.config``
- :class:`nbfigtulz.figtools.FigContext` as ``nbfigtulz.FigContext``
- :class:`nbfigtulz.figtools.with_context` as ``nbfigtulz.with_context``
- :class:`nbfigtulz.figtools.save_fig` as ``nbfigtulz.save_fig``
- :class:`nbfigtulz.figtools.img_grid` as ``nbfigtulz.img_grid``
- :class:`nbfigtulz.figtools.Size` as ``nbfigtulz.fig_size``
"""

from .config import config
from .figtools import FigContext, with_context, save_fig, img_grid
from .figtools import Size as fig_size